/****************************************************************************
 This is the Automatic Photos module for the Roxen WebServer 2.2
 Copyright (C) 2001,2002 -- M Norrby <norrby@lysator.liu.se>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         
 
 To acquire the latest version of aphoto, check out 
 <URL:http://www.lysator.liu.se/~norrby/aphoto.html>.
 
 Please send comments and bug-reports to <norrby@lysator.liu.se>.
****************************************************************************/

#include <module.h>
#include <request_trace.h>
inherit "module";

static roxen.ImageCache the_cache;

constant cvs_version
= "$Id: aphoto.pike,v 1.102 2003/01/05 12:42:59 norrby Exp $";
constant module_type = MODULE_LOCATION|MODULE_TAG;
constant module_name = "Automatic photos";
constant thread_safe = 1;
constant module_doc = 
"A photo album module which generates its own thumbnails and "
"has medium sized images with descriptions for browsing.";

static string all_but_aphoto_text_in_jpeg_comment(string file)
{
  return jpeg_comment(file) - 
    (make_xml(MODULE_NAME, (aphoto_top_tag(file) || "")));
}

static mapping aphoto_error(string whatHappened, RequestID id)
{
  string page = PAGE_HEADER +    
    "<head>\n <title>"
    "Aphoto server module error"
    "</title>\n</head>\n"
    
    "<body>\n <h1>"
    "Aphoto server module error!"
    "</h1>\n"
    " <p>" + whatHappened + "</p>\n</body>\n" +
    PAGE_FOOTER;
  return Roxen.http_low_answer(500,
			       Roxen.parse_rxml(page, id));
}

static string aphoto_top_tag(string filename)
{
  if (filename == cache_last_aphoto_top_tag_filename) {
    return cache_last_aphoto_top_tag;
  }
  cache_last_aphoto_top_tag = extract_xml(MODULE_NAME, jpeg_comment(filename));
  cache_last_aphoto_top_tag_filename = filename;
  return aphoto_top_tag(filename);
}

static string cache_last_aphoto_top_tag;
static string cache_last_aphoto_top_tag_filename;
static string cache_last_dir_info_top_tag;
static string cache_last_dir_info_top_tag_dir;
static mapping(string:array(int)|Stat) cache_last_dir_stat;
static string cache_last_dir_stat_path;

static string combine_url(string begin, string end)
{
  if (!sizeof(begin) || !sizeof(end))
    return begin + end;
  return begin[..sizeof(begin) - 2]
    + combine_path(begin[sizeof(begin) - 2..sizeof(begin) - 2], end[0..0])
    + end[1..];
}

void create(Configuration conf)
{
  roxen.add_permission(DATA_EDITING_PERMISSION, DATA_EDITING_PERMISSION);
  set_module_creator("norrby@lysator.liu.se");
  set_module_url(MODULE_URL);
  defvar("location",
	 Variable.Location("/photoalbum/",
			   0,
			   "Mount point",
			   "This is where the module will be inserted "
			   "in the namespace of your server."));
  defvar("albumName",
	 Variable.String("My album",
			 0,
			 "Photo album name",
			 "The name of your photo album. It will set the "
			 "variable &amp;form.albumName; for you."));

  defvar("thumbnailSize",
	 Variable.Int(128,
		      0,
		      "Thumbnail size",
		      "Size of automatically generated thumbnails."));

  defvar("makeMidSizeImages",
	 Variable.Flag(1,
		       0,
		       "Generate medium sized images",
		       "If the module should create medium sized images "
		       "for browsing. Useful, since photos often are "
		       "impractically large. Not recommended for slow "
		       "computers with a small amount of memory."));

  defvar("midImageSize",
	 Variable.Int(640,
		      0,
		      "Medium sized image size",
		      "Size of the browseable pictures which "
		      "have a previous and a next button. This setting "
		      "only has effect when images are larger than this "
		      "value, and \"Generate medium sized images\" is "
		      "activated!"));
  getvar("midImageSize")->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
				    {
				      return query("makeMidSizeImages") == 0;
				    });	 
  defvar("makeDropShadow",
	 Variable.Flag(1,
		       0,
		       "Generate drop shadow effects",
		       "If the module should create drop shadows "
		       "around both thumbnails and mid sized images."));

  defvar("thumbnailPageSource",
	 Variable.StringChoice("default", 
			       (["default":"Use default page",
				 "file":"Use my own file",
				 "text":"Define page in config interface"]),
			       0,
			       "Thumbnail page source",
			       "From where should aphoto fetch the page "
			       "template for the thumbnail page?"));

  defvar("thumbnailPage",
	 Variable.String("thumb.html",
			 0,
			 "Thumbnail page",
			 "Filename of the page to be used as template for the "
			 "thumbnail pages. The path could be absolute, or "
			 "realtive the directory containing the photo tree. "));
  getvar("thumbnailPage")->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
				    {
				      return query("thumbnailPageSource")
					!= "file";
				    });

  defvar("photoPageSource",
	 Variable.StringChoice("default", 
			       (["default":"Use default page",
				 "file":"Use my own file",
				 "text":"Define page in config interface"]),
			       0,
			       "Photo page source",
			       "From where should aphoto fetch the page "
			       "template for the medium sized photo page?"));
  defvar("thumbnailPageText",
	 Variable.Text(DEFAULT_THUMBNAIL_PAGE,
		       0,
		       "Custom thumbnails page",
		       "Config interface defined custom thumbnails page HTML"));
  getvar("thumbnailPageText")->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
				    {
				      return
					query("thumbnailPageSource") != "text";
				    });


  defvar("photoPage",
	 Variable.String("photo.html",
			 0,
			 "Photo page",
			 "Filename of the page to be used as template for the "
			 "photo pages. The path could be absolute, or realtive "
			 "the directory containing the photo tree."));
  getvar("photoPage")->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
				    {
				      return query("photoPageSource") != "file";
				    });

  defvar("photoPageText",
	 Variable.Text(DEFAULT_PHOTO_PAGE,
		       0,
		       "Custom photo page",
		       "Config interface defined custom photo page HTML"));
  getvar("photoPageText")->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
				    {
				      return query("photoPageSource") != "text";
				    });

  defvar("rdjpgcomPath",
	 Variable.File("/usr/bin/rdjpgcom",
		       0,
		       "Path to JPEG comment extractor",
		       "Because of some shortcomings in Pike, the program "
		       "<b>rdjpgcom(1)</b> is used to extract image comment "
		       "information where image descriptions are stored. "
		       "This variable needs to have the full path to this "
		       "program! If you are using Debian, the package to "
		       "install is <b>libjpeg-progs</b>, and if you are "
		       "using RedHat, the package name is libjpeg."));
	 
  defvar("wrjpgcomPath",
	 Variable.File("/usr/bin/wrjpgcom",
		       0,
		       "Path to JPEG comment writer",
		       "Because of some shortcomings in Pike, the program "
		       "<b>wrjpgcom(1)</b> is used to write image comment "
		       "information. This variable needs to have the full "
		       "path to this program! If you are using Debian, the "
		       "package to install is <b>libjpeg-progs</b>, and if "
		       "you are using RedHat the package name is libjpeg."));

  defvar("jpegtranPath",
	 Variable.File("/usr/bin/jpegtran",
		       0,
		       "Path to JPEG lossless transformation utility",
		       "This variable needs to have the full "
		       "path to this program! If you are using Debian, the "
		       "package to install is <b>libjpeg-progs</b>, and if "
		       "you are using RedHat the package name is libjpeg."));
			 
	 
  defvar("LC_ALL",
	 Variable.String("sv_SE.iso8859-1",
			 0,
			 "Locale setting",
			 "Aphoto will set LC_ALL to this value before running "
			 "the <b>wrjpgcom</b> and <b>rdjpgcom</b> programs."));
  defvar("hideFiles",
	 Variable.StringList(({".www_not_browsable",
			       ".nodiraccess",
			       ".hidden",}),
			     0,
			     "Files that toggle visibility",
			     "If a file with any of these filenames exists in "
			     "a directory, the actual directory will not be "
			     "shown in "
			     "listings, and it can not be browsed. You can "
			     "still reach the files directly though. Useful "
			     "if you want private directories for button "
			     "graphics and such in your web pages."));
	 
  defvar("rootDir",
	 Variable.Directory("/usr/local/photos",
			    0,
			    "Directory containing photos",
			    "Directory in real file system "
			    "containing the photo tree."));

  defvar("dirInfoFilename",
	 Variable.String(".aphoto",
			 0,
			 "Dirinfo filename",
			 "Filename of info file used to store directory "
			 "information and per directory sorting options."));

  defvar("authMethod",
         Variable.StringChoice("configInterfaceUsers",
			       ([ "configInterfaceUsers":"Administration "
				  "Interface User",
				  "authModule":"Authentification Module" ]),
			       0,
			       "Authentication method",
			       "The simplest way to "
			       "get aphoto to work is to use administration "
			       "interface"
			       "users. For example you can use the default"
			       "administrator user, which has all permissions."
			       "Otherwise you could create a new user with only"
			       "Aphoto data editing permissions in the Roxen"
			       "config interface."));

  defvar("secureServerRedirection",
	 Variable.Flag(0,
		       0,
		       "Secure server redirection",
		       "Should aphoto redirect you to another server/port "
		       "when editing image descriptions? "
		       "(This option requires at least Roxen 2.2.291)"));

  defvar("secureServerRedirectionHost",
	 Variable.URL("https://your.aphoto.server.foo/",
		      0,
		      "Secure server redirection host and protocol",
		      "Where should we redirect to, in case you answered "
		      "yes to the <b>Secure server redirection</b>?"));
  getvar("secureServerRedirectionHost")->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
				    {
				      return query("secureServerRedirection")
					== 0;
				    });
  
  defvar("imageSortOrder",
	 Variable.StringChoice("filename",
			       ([ "mtime":"Last modified",
				  "filename":"Filename", 
			          "-filename":"Filename, reverse",
				  "-title":"Image title (or short description)",
			       ]),
			       0,
			       "Default sort order for images",
			       "The sort order to use by default for "
			       "thumbnails and images"));

  defvar("dirSortOrder",
	 Variable.StringChoice("-filename",
			       ([ "mtime":"Last modified",
				  "filename":"Directory name in filesystem, "
				  "lowest first",
				  "-filename":"Directory name in filesystem, "
				  "highest first",
				  "infoorfilename":"Folder title (short "
				  "description)"
			       ]),
			       0,
			       "Default sort order for directories"));
  defvar("suffix",
	 Variable.String(".html",
			 0,
			 "Suffix of links to pictures",
			 "When producing a link to a particular picture, "
			 "this suffix will be added, e g: foo.jpeg.html. " 
			 "If no suffix is given, thumbnails will link "
			 "directly to the images, but then (of course) "
			 "you will not get any browse buttons."));

  defvar("indexFileName",
	 Variable.String("index.html",
			 0,
			 "Index file override file name",
			 "Name of index file if overriding of \"Thumbnail "
			 "page\" is desired in a particular directory. To "
			 "override the \"Thumbnail page\" in a directory, "
			 "just put a file named like the \"Index file "
			 "override file name\" in that directory."));
}

static mapping dir_edit(string dirInfoFilename, RequestID id)
{
  id->variables->full_path = dirInfoFilename;
  return Roxen.http_rxml_answer(DEFAULT_THUMBNAIL_EDIT_PAGE, id);
}

static int dir_date(string dir)
{
  return (int) (extract_xml(DIR_DATE, dir_info_top_tag(dir))||0);
}

static string dir_info(string file) {
  return Stdio.read_bytes(file);
}

static string dir_info_long(string dir)
{
  return extract_xml(DIR_INFO_LONG, dir_info_top_tag(dir));
}

static string dir_info_short(string dir)
{
  return extract_xml(DIR_INFO_SHORT, dir_info_top_tag(dir));
}

static string dir_info_top_tag(string dir)
{
  if (dir == cache_last_dir_info_top_tag_dir) {
    return cache_last_dir_info_top_tag;
  }
  cache_last_dir_info_top_tag_dir = dir;
  cache_last_dir_info_top_tag =
    extract_xml(MODULE_NAME,
		dir_info(combine_path(dir, query("dirInfoFilename"))));
  return dir_info_top_tag(dir);
}
static array dir_sort(string path,
		      array dirs,
		      mapping fileStatsInDir,
		      RequestID id)
{
  int(0..1) mtime_sort_fun(string left, string right) {
    return fileStatsInDir[left]->mtime > fileStatsInDir[right]->mtime;
  };
  
  int(0..1) dirinfo_sort_fun(string left, string right) {
    string infoLeft = dir_info_short(combine_path(path, left)) || left;
    string infoRight = dir_info_short(combine_path(path, right)) || right;
    if (infoLeft == "") infoLeft = left;
    if (infoRight == "") infoRight = right;
    return infoLeft > infoRight;
  };

  switch (query("dirSortOrder")) {
  case "mtime":
    return Array.sort_array(dirs, mtime_sort_fun);
  case "filename":
    return Array.sort_array(dirs, filename_sort_fun);
  case "-filename":
    return Array.sort_array(dirs, rev_filename_sort_fun);
  case "infoorfilename":
    return Array.sort_array(dirs, dirinfo_sort_fun);
  };
  return 0;
}

static mapping directory_page(RequestID id)
{
  string page = "";
  switch (query("thumbnailPageSource")) {
  case "default":
    page = DEFAULT_THUMBNAIL_PAGE;
    break;
  case "text":
    page = query("thumbnailPageText");
    break;
  case "file":
    page = Stdio.read_bytes(query("thumbnailPage")) ||
      Stdio.read_bytes(combine_path(query("rootDir"), query("thumbnailPage")));
    if (!page) {
      return aphoto_error("Couldn't find your custom thumbnail page \""
			  + query("thumbnailPage") + "\"", id);
    }
    break;
  default:
    return aphoto_error("Error in thumbnail page configuration!", id);
  };
  return Roxen.http_rxml_answer(page, id);
}

static mapping dir_stat(string path, RequestID id)
{ 
  mapping a = id->conf->find_dir_stat(query("location") + dirname(path)
				      + "/", id);    
  if (!a || !sizeof(a))
    return 0;

  return a;
}

static int(0..1) editing_data_posted(RequestID id)
{
  return (/*id->variables->edit && */id->variables->aphoto_editing_data);
}

static int(0..1) in_edit_mode(RequestID id)
{
  return !!id->variables->edit && id->variables->edit == "true";
}

static int exif_time(string jpegFilename)
{
  mapping exif = Standards.EXIF.get_properties(Stdio.FILE(jpegFilename));
  if (sizeof(exif)) {
    if (!exif->DateTime)
      return 0;
    return unixtime(exif->DateTime);
  }
  return 0;
}

static string|int(0..1) extract_xml(string tag, string textBlock)
{
  if (!textBlock) return 0;
  if (sscanf(textBlock, "%*s<"+tag+"%*s>%s</"+tag+">",
	     textBlock) != 3)
    return sscanf(textBlock, "%*s<"+tag+"%*s/>") >= 1;
  return textBlock;
}

static mapping file_edit(string imagename, string fullPath, RequestID id)
{
  id->variables->full_path = fullPath;
  if (image_hidden(fullPath)) {
    id->variables->hidden = "yes";
  }
  return Roxen.http_rxml_answer(DEFAULT_PHOTO_EDIT_PAGE, id);  
}

static void flush_cache() {
  the_cache->flush();
}

static void flush_cache_dir_info_top_tag()
{
  cache_last_dir_info_top_tag_dir = 0;
}

static void flush_request_local_caches()
{
  flush_cache_last_dir_stat();
  flush_cache_dir_info_top_tag();
  flush_cache_aphoto_top_tag();
}

static int(0..1) file_is_image(string filename, RequestID id)
{
  return has_prefix(id->conf->type_from_filename(filename), "image/");
}

static int(0..1) file_is_video(string filename, RequestID id)
{
  return has_prefix(id->conf->type_from_filename(filename), "video/");
}

static int(0..1) filename_sort_fun(string left, string right) {
  return left > right;
};

array(string)|void find_dir(string path, RequestID id)
{
  string file = combine_path(query("rootDir"), remove_first_slash(path));
  return get_dir(file);
}

mapping(string:array(int)|Stat) find_dir_stat(string path, RequestID id)
{
  if (cache_last_dir_stat_path == path) {
    return cache_last_dir_stat;
  }
  
  array(string) files = find_dir(path, id);
  mapping(string:Stat) res = ([]);

  foreach(files || ({}), string fname) {
    Stat st = stat_file(combine_path(path, fname), id);
    if (st) {
      res[fname] = st;
    }
  }

  (cache_last_dir_stat = res) && (cache_last_dir_stat_path = path);
  return res;
}

mapping|Stdio.File|void find_file(string path, RequestID id)
{
  TRACE_ENTER("aphoto find_file",0);
  flush_request_local_caches();

  if (should_we_redirect_to_secure_server(id)) {
    TRACE_LEAVE("Redirect");
    return Roxen.http_redirect(combine_url(query("secureServerRedirectionHost"),
					   id->not_query)
			       + query_part_of_url(id));
  }

  if (permission_denied(id)) {
    TRACE_LEAVE("Permission denied");
    return id->conf->authenticate_throw(id, DATA_EDITING_PERMISSION);
  }

  if (editing_data_posted(id)) {
    TRACE_LEAVE("Editing post");
    return handle_editing_post(id);
  }

  id->variables->internal = query_absolute_internal_location(id);
  id->variables->aphotoversion = cvs_version[19..23];

  string filename = real_file(path, id);
  Stat s = file_stat(filename);
  if (!s) {
    mapping ret = show_not_real_file(filename, path, id);
    TRACE_LEAVE("Generated file");
    return ret;
  }

  if (s->isdir) {
    mapping ret = show_directory(filename, id);
    TRACE_LEAVE("Directory");
    return ret;
  }

  if (is_jpeg_file(filename) && image_hidden(filename) &&
      !in_edit_mode(id)) {
    TRACE_LEAVE("Hidden file");
    return 0;
  }
  TRACE_LEAVE("Real file");
  return Stdio.File(filename);
}

mapping|int(-1..0)|Stdio.File find_internal(string file, RequestID id)
{
  switch (basename(file)) {
  case "aphoto.css":
    return Roxen.http_string_answer(APHOTO_CSS, "text/css");
    break;
  case "arrow-right":
    return Roxen.http_string_answer(ARROW_RIGHT, "image/gif");
    break;
  case "arrow-up":
    return Roxen.http_string_answer(ARROW_UP, "image/gif");
    break;
  case "arrow-left":
    return Roxen.http_string_answer(ARROW_LEFT, "image/gif");
    break;
  case "arrow-right-nonactive":
    return Roxen.http_string_answer(ARROW_RIGHT_NONACTIVE, "image/gif");
    break;
  case "arrow-up-nonactive":
    return Roxen.http_string_answer(ARROW_UP_NONACTIVE, "image/gif");
    break;
  case "arrow-left-nonactive":
    return Roxen.http_string_answer(ARROW_LEFT_NONACTIVE, "image/gif");
    break;
  case "folder":
    return Roxen.http_string_answer(FOLDER, "image/gif");
    break;
  }
  return the_cache->http_file_answer((file/".")[0], id);
}

static array(string) find_prev_next_dir(string path, string dir, RequestID id)
{
  mapping(string:array(int)|Stat) stat = id->conf->find_dir_stat(path, id);
  
  if (!stat || !sizeof(stat)) {
    return ({0, 0, 0, 0});
  }
  array(string) dirs = remove_nondir_and_hidden(stat, path, id);
  if (!sizeof(dirs)) {
    return ({0, 0, 0, 0});
  }
  dirs = dir_sort(id->conf->real_file(path, id), dirs, stat, id);

  string prev = 0, next = 0;

  string first = dirs[0];
  string last = dirs[-1];
  
  for (int i = 0; i < sizeof(dirs); ++i) {
    if (i > 0 && dirs[i] == dir) {
      prev = dirs[i - 1];
    }
    if (i < sizeof(dirs) - 1 && dirs[i] == dir)
      next = dirs[i + 1];
  }
  return ({ prev, next, first, last });
}

static void flush_cache_aphoto_top_tag()
{
  cache_last_aphoto_top_tag_filename = 0;
}

static void flush_cache_last_dir_stat()
{
  cache_last_dir_stat_path = 0;
}

static array(Image.Layer) generate_image(mapping args, RequestID id)
{
  array layers = ({});
  mapping opts = ([]);
  
  Image.Image image;
  if (file_is_video(args->file, id) && supports_video_decoding()) { 
#if constant(_Ffmpeg.CODEC_ID_MJPEG)
    _Ffmpeg.ffmpeg decoder = _Ffmpeg.ffmpeg(_Ffmpeg.CODEC_ID_MJPEG, 0);
    mapping decodeMapping = decoder->decode(Stdio.read_bytes(args->file));
    image = Image._decode(decodeMapping->data)->img;
#endif
  } else {
    image = Image._decode(Stdio.read_bytes(args->file))->img;
  }
  
  int(0..) xSize = 0;
  int(0..) ySize = 0;
  int(0..1) isThumbnail = (args["imagetype"] == "thumbnail") ? 1 : 0;
  int(0..1) isHidden = image_hidden(args->file);
  if (isThumbnail) {
    xSize = query("thumbnailSize");
    ySize = xSize;
  } else {
    xSize = query("midImageSize");
    ySize = xSize;
  }
  if (image->xsize() > image->ysize()) {
    ySize = 0;
  } else {
    xSize = 0;
  }
  if (image->xsize() < xSize) xSize=0;
  if (image->ysize() < ySize) ySize=0;
  if (xSize||ySize) image = image->scale(xSize, ySize);
  if (query("makeDropShadow")) {
    image = make_dropshadow(image,
			    image->xsize(),
			    isThumbnail ? THUMBNAIL_PASSEPARTOUT_SIZE
			    : PHOTO_PASSEPARTOUT_SIZE,
			    SHADOW_BORDER_TOPLEFT,
			    SHADOW_BORDER_BOTTOMRIGHT);
  }
  if (isHidden) {
    image = image->grey(0,0,255);
  }
  return ({Image.Layer(image)});
}

static mapping get_my_args(mapping args, RequestID id)
{
  if (args->file)
    catch 
    {
      Stat st = file_stat(args->file);
      if (st)
	{
	  Roxen.add_cache_stat_callback(id, args->file, st->mtime);
	  args->mtime = (string) (args->stat = st->mtime);
	  args->filesize = (string) st->size;
	}
    };

  return args;
}

static mapping handle_dir_editing_post(string filename, RequestID id)
{
  Stdio.File dirInfoFile = Stdio.File(filename, "wtc");
  dirInfoFile->write(make_dir_info((id->variables->new_dir_info_short != 
				    id->variables->short_dir_path) ? 
				   id->variables->new_dir_info_short : "",
				   id->variables->new_dir_info_long,
				   (string) unixtime(id->variables->
						     new_date)));
  dirInfoFile->close();
  return Roxen.http_redirect(id->not_query);
}

static mapping handle_editing_post(RequestID id)
{
  string filename = id->variables->full_path_in_real_file_system;
  if (basename(filename) == query("dirInfoFilename")) {
    return handle_dir_editing_post(filename, id);
  } else {
    return handle_photo_editing_post(filename, id);
  }
}

static mapping handle_photo_editing_post(string filename, RequestID id)
{
  string oldFilename = filename + ".old";
  string newFilename = filename + ".new";
  string comment = make_jpeg_comment(filename,
				     id->variables->new_image_desc_short,
				     id->variables->new_image_desc_long,
				     (string) unixtime(id->variables->
						       new_date),
				     !!id->variables->hide_and_next);
  mixed comment_error =
    write_new_jpeg_comment(filename, oldFilename, newFilename, comment, id);
  if (comment_error) return comment_error;

  if (images_equal(filename, newFilename)) {
    swap_files(filename, oldFilename, newFilename);
  } else {
    return aphoto_error("Something went very wrong in writing "
			"a JPEG comment in file. Exiting. Your"
			"file was not damaged.", id);
  }

  if (images_equal(filename, oldFilename)) {
    rm(oldFilename);
  } else {
    return aphoto_error("Something went very wrong.\n"
			"Old file not deleted, but resides in."
			+ oldFilename, id);
  }

  mixed transform_error =
    transform_image(filename, oldFilename, newFilename, id);
  if (transform_error) return transform_error;

  if (!!id->variables->submit_and_previous &&
      !!id->variables->previous &&
      id->variables->previous != "")
    {
      return Roxen.http_redirect(id->variables->previous);
    }
  if ((!!id->variables->submit_and_next ||
       !!id->variables->hide_and_next) &&
      !!id->variables->next &&
      id->variables->next != "") {
    return Roxen.http_redirect(id->variables->next);
  }
  if (!!id->variables->hide_and_next || !!id->variables->unhide) {
    return Roxen.http_redirect(id->not_query + query_part_of_url(id));
  }
  return Roxen.http_redirect(id->not_query);
}

static int(0..1) runs_ssl(RequestID id) {
  return id->my_fd
    && id->my_fd->session
    && id->my_fd->session->cipher_spec->key_bits;
}

static int image_date(string file)
{
  return (int) extract_xml(IMAGE_DATE, aphoto_top_tag(file));
}

static int(0..1) image_hidden(string file)
{
  return extract_xml(IMAGE_HIDDEN, aphoto_top_tag(file));
}

static string image_desc_long(string file)
{
  return extract_xml(IMAGE_DESC_LONG, aphoto_top_tag(file));
}

static string image_desc_short(string file)
{
  return extract_xml(IMAGE_DESC_SHORT, aphoto_top_tag(file));
}

static int(0..1) images_equal(string imageFilename1, string imageFilename2)
{
  return Image.JPEG.decode(Stdio.read_bytes(imageFilename1)) == 
    Image.JPEG.decode(Stdio.read_bytes(imageFilename2));
}

static int(0..1) image_file_exists(string filename)
{
  if (file_stat(filename)) {
    return 1;
  }
  return 0;
}

static int(0..1) image_file_ok(string imageFilename)
{
  if (!file_stat(imageFilename)->isreg) {
    return 0;
  }
  mixed error = catch {
    Image.JPEG.decode(Stdio.read_bytes(imageFilename));
  };
  if (error) {
    return 0;
  }
  return 1;
}

static array image_sort(string path,
			array files,
			mapping fileStatsInDir,
			RequestID id)
{

  function contenttype = id->conf->type_from_filename;
  files = filter(files, lambda(string fn) {
			  return has_prefix(contenttype(fn), "image/") ||
			    supports_video_decoding()
			    && has_prefix(contenttype(fn), "video/");
			});

  int(0..1) mtime_sort_fun(string left, string right) {
    return fileStatsInDir[left]->mtime > fileStatsInDir[right]->mtime;
  };
  
  int(0..1) title_sort_fun(string left, string right) {
    left = image_desc_short(combine_path(path, left)) || left;
    right = image_desc_short(combine_path(path, right)) || right;
    return left > right;    
  };
  
  switch (query("imageSortOrder")) {
  case "mtime":
    return Array.sort_array(files, mtime_sort_fun);
  case "filename":
    return Array.sort_array(files, filename_sort_fun);
  case "-filename":
    return Array.sort_array(files, rev_filename_sort_fun);
  case "-title":
    return Array.sort_array(files, title_sort_fun);
  };
  return 0;
}

static int(0..1) is_jpeg_file(string filename)
{
  if (!file_stat(filename) || !file_stat(filename)->isreg
      ||file_stat(filename)->size < 2) {
    return 0;
  }
  string idString = Stdio.read_bytes(filename, 0, 2);
  return idString == "\377\330";
}

static int(0..1) is_old_netscape(RequestID id)
{
  array client = id->client;
  if (client && sizeof(client)) {
    return client[0][0..8] == "Mozilla/4";
  }
}

static string jpeg_comment(string file)
{
  if (!is_jpeg_file(file)) {
    return 0;
  }
  if (supports_jpeg_transforms()) {
    string ret = Image.JPEG.decode_header(Stdio.read_bytes(file))->comment;
    return ret || "";
  } else {
    Stdio.File in = Stdio.File();
    mixed error = catch {
      Process.create_process(({ query("rdjpgcomPath"), file }),
			     ([ "env":getenv() + (["LC_ALL":query("LC_ALL")]),
				"stdout":in->pipe() ]));
    };
    if (error) {
      return 0;
    }
    string ret = in->read();
    if (ret != "" && ret[-1] == '\n')
      return ret[..sizeof(ret)-2];
    else
      return ret;
  }
}

static string make_dir_info(string dirInfoShort,
			    string dirInfoLong,
			    string dirDate)
{
  return make_xml(MODULE_NAME,
		  "\n" + make_xml(DIR_INFO_SHORT, dirInfoShort) + "\n" +
		  make_xml(DIR_INFO_LONG, dirInfoLong) + "\n" +
		  make_xml(DIR_DATE, dirDate) + "\n");
}

static string make_jpeg_comment(string filename,
				string imageDescShort,
				string imageDescLong,
				string imageDate,
				int(0..1) imageHidden)
{
  return all_but_aphoto_text_in_jpeg_comment(filename) +
    make_xml(MODULE_NAME,
	     "\n" + make_xml(IMAGE_DESC_SHORT, imageDescShort) + "\n" +
	     make_xml(IMAGE_DESC_LONG, imageDescLong) + "\n" +
	     make_xml(IMAGE_DATE, imageDate) + "\n" +
	     (imageHidden ? make_xml(IMAGE_HIDDEN, 0) : ""));
}

static void make_mid_size_image(string imagename, string filename, RequestID id)
{
  string midSizeImageOrImageItself = mid_image_name(imagename, filename, id);
  id->variables->midiphoto = midSizeImageOrImageItself;
  string path = id->conf->real_file(imagename, id);
  int(0..) size = Stdio.file_size(path); 
  id->variables->fullimagesizebytes = (string) size;
  id->variables->fullimagesizekbytes = (string) (size/1024);
  id->variables->fullimagesizembytes = (string) (size/1024/1024);
  Stdio.File fd = open(path, "r");
  int(0..) fullX, fullY, midX, midY;
  [fullX,fullY] = Dims.dims()->get(fd) || ({0,0});
  fd->close();
  id->variables->fullimagewidth = (string) fullX;
  id->variables->fullimageheight = (string) fullY;
  [midX, midY] = mid_image_size(fullX, fullY);
  id->variables->midiphotowidth = (string) midX;
  id->variables->midiphotoheight = (string) midY;
}

static string make_xml(string tag, string textBlock)
{
  if (textBlock != 0) {
    return "<" + tag + ">" + textBlock + "</" + tag + ">";
  } else {
    return "<" + tag + "/>";
  }
}

static mapping mid_image_page(RequestID id)
{
  string page = "";
  switch (query("photoPageSource")) {
  case "default":
    page = DEFAULT_PHOTO_PAGE;
    break;
  case "text":
    page = query("photoPageText");
    break;
  case "file":
    page = Stdio.read_bytes(query("photoPage")) ||
      Stdio.read_bytes(combine_path(query("rootDir"), query("photoPage")));
    if (!page) {
      return aphoto_error("Couldn't find your custom photo page \""
			  + query("photoPage") + "\"", id);
    }
    break;
  default:
    return aphoto_error("Error in photo page configuration!", id);
  };
  return Roxen.http_rxml_answer(page, id);
}

static string mid_image_name(string imagename, string filename, RequestID id)
{
  string midSizeImageOrImageItself;
  if (query("makeMidSizeImages")) {
    midSizeImageOrImageItself = query_absolute_internal_location(id)
      + the_cache->store(get_my_args(([
	"imagetype":"midimage",
	"file":filename,
	"query":query_part_of_url(id),
	"format":"jpg",
      ]), id), id); 
  } else {
    midSizeImageOrImageItself = imagename;
  }
  return midSizeImageOrImageItself;
}

static array(int(0..)) mid_image_size(int(0..) xSizeOfFullImage,
				      int(0..) ySizeOfFullImage)
{
  if (!query("makeMidSizeImages") || xSizeOfFullImage == 0 
      || ySizeOfFullImage == 0) {
    return ({ xSizeOfFullImage, ySizeOfFullImage });
  }
    
  float scale = (float) query("midImageSize") / 
    max(xSizeOfFullImage, ySizeOfFullImage);
  if (scale > 1) {
    scale = 1.0;
  }
  int border = shadow_borders(PHOTO_PASSEPARTOUT_SIZE);
  return ({ border + (int)(xSizeOfFullImage*scale),
	    border + (int)(ySizeOfFullImage*scale) });
}

static string parent(string dir, RequestID id) {
  string realfile = id->conf->real_file(dir, id);
  if (file_stat(realfile)->isdir)
    if (dir[-1] != '/') {
      dir += "/";
    }
  string result = dirname(dirname(dir)) + "/";
  if (sizeof(dirname(dir)) <= sizeof(query("location"))) {
    result = "";
  } 
  return result;
}

static int(0..1) permission_denied(RequestID id)
{
  if (in_edit_mode(id))
    {
      string actualUser = 0;
      if (query("authMethod") == "configInterfaceUsers") {
        actualUser = roxen.configuration_authenticate(id,
						      DATA_EDITING_PERMISSION);
      } else {
	User user = id->conf->authenticate(id);
  	if (user) {
	  return 0;
	}

      }
      if (actualUser == "0" || actualUser == 0) {
	return 1;
      } else {
	return 0;
      }
    }
  return 0;
}

static string program_status(Variable.File file)
{
  if (!!file->stat() && !file->stat()->isdir)
    return
      "<img src=\"/internal-roxen-err_1\" alt=\"!\" />"
      "<code>" + file->query() +
      "</code> seems to exist. Good.<br />\n";
  
  return
    "<img src=\"/internal-roxen-err_3\" alt=\"!\" />"
    "<code>" + file->query() +
    "</code> does not seem to exist!<br />\n";
}

mapping(string:function) query_action_buttons() {
  return ([ "Clear image cache":flush_cache ]);
}

static string query_part_of_url(RequestID id)
{
  return id->query ? "?" + id->query : "";
}

string|void real_file(string path, RequestID id)
{
  string file = combine_path(query("rootDir"), remove_first_slash(path));
  Stat st = file_stat(file);
  if (!st)
    return(file);
  if (st->isdir) {
    if (file[-1] != '/') {
      file += "/";
    }
    return(file);
  }
  return(file);
}

static string remove_first_slash(string path)
{
  if (sizeof(path) && path[0] == '/')
    return path[1..];
  return path;
}

static string remove_last_slash(string path)
{
  if (sizeof(path) == 1)
    return (path[0] == '/' ? "" : path);
  if (path != "" && path[-1] == '/')
    return path[..sizeof(path) - 2];
  return path;
}

array(string) remove_hidden_images(array(string) files, string path)
{
  foreach(files, string f) {
    if (is_jpeg_file(combine_path(path, f)) &&
	image_hidden(combine_path(path, f))) {
      files -= ({ f });
    }
  }
  return files;
}

array(string) remove_nondir_and_hidden(mapping a, string d, RequestID id)
{
  array(string) files = indices(a);
  array tmp = ({});
  int hide = 0;
  array hideFiles = query("hideFiles");
  foreach(hideFiles, string hideFile)
    if (file_stat(id->conf->real_file
(combine_path(combine_path(d,
								hideFile)),
				      id)))
      hide = 1;
  if (!hide) {
    foreach(files, string filename) {
      if (a[filename]->isdir) {
	// check if directory should be hidden
	int hide = 0;
	foreach (hideFiles, string hideFile) {
	  string realfile = 
	    id->conf->real_file(combine_path(combine_path(d,filename),
					     hideFile), id);
	  if (file_stat(realfile)) {
	    hide = 1;
	  }
	}
	if (!hide) {
	  tmp += ({filename});
	}
      }
    }
  }
  files = tmp;
  return files;
}

static string remove_suffix(string filename)
{
  string sfx = query("suffix");
  if(!has_suffix(filename, sfx)) return filename;
  return filename[..sizeof(filename) - sizeof(sfx) - 1];
}

static int(0..1) rev_filename_sort_fun(string left, string right) {
  return left < right;
};
  
static void set_previous_and_next_variables(array files,
					    string path,
					    RequestID id)
{
  for (int j = 0; j < sizeof(files); j++) {
    if (combine_path(dirname(path), files[j]) + query("suffix") == path) {
      id->variables->photofilename = files[j];
      if (j != 0) {
	string imagename = combine_path(query("location") + dirname(path) + "/",
					files[j-1]);
	string filename = combine_path(query("rootDir"),
				       combine_path(dirname(path), files[j-1]));
	id->variables->prefetchleft = mid_image_name(imagename, filename, id);
	id->variables->previous = imagename
	  + query("suffix") + query_part_of_url(id);
      }
      if (j < (sizeof(files) - 1)) {
	string imagename = combine_path(query("location") + dirname(path) + "/",
					files[j+1]);
	string filename = combine_path(query("rootDir"),
				       combine_path(dirname(path), files[j+1]));
	id->variables->prefetchright = mid_image_name(imagename, filename, id);
	id->variables->next =  imagename
	  + query("suffix") + query_part_of_url(id);
      }
    }
  }
  id->variables->parent = "./"  + query_part_of_url(id);
  id->variables->start = query("location");
  id->variables->first = combine_path(query("location")
				      + dirname(path) + "/", files[0])
    + query("suffix") + query_part_of_url(id);
  id->variables->last = combine_path(query("location")
				     + dirname(path) + "/", files[-1])
    + query("suffix") + query_part_of_url(id);
}

static int shadow_borders(int passepartout)
{
  return query("makeDropShadow") * (2 * passepartout
				    + SHADOW_BORDER_TOPLEFT
				    + SHADOW_BORDER_BOTTOMRIGHT);
}

void start(int num, Configuration config)
{
  the_cache = roxen.ImageCache("aphoto", generate_image);

  array dependencies = ({"contenttypes",
			 "rxmlparse",
			 "rxmltags",
                         "auth_httpbasic"});
  if (query("authMethod") == "authModule") {
    dependencies += ({"userdb_system"});
  }

  module_dependencies (config, dependencies);
}

static mapping show_directory(string filename, RequestID id)
{
  string f = id->not_query;
  if (f == "" || (f[-1] != '/' && f[-1] != '.'))
    return Roxen.http_redirect(f + "/" + query_part_of_url(id), id);
  string p = parent(id->not_query, id);
  if (sizeof(p) > 0) {
    id->variables->parent = parent(id->not_query, id) + query_part_of_url(id);
    id->variables->start = query("location");
    string prev = 0, next = 0, first = 0, last = 0;
    [prev, next, first, last] = 
      find_prev_next_dir(p,
			 basename(remove_last_slash(filename)),
			 id);
    if (prev) {
      id->variables->previous = combine_url(id->variables->parent,
					    "/" + prev + "/");
      id->variables->first = combine_url(id->variables->parent,
					 "/" + first + "/");
    }
    if (next) {
      id->variables->next = combine_url(id->variables->parent,
					"/" + next + "/");
      id->variables->last = combine_url(id->variables->parent,
					"/" + last + "/");
      
    }
  }

  id->variables->albumname = query("albumName");
  id->variables->longdirinfo = dir_info_long(filename) || "";
  id->variables->datelong = stringtime((int) (dir_date(filename) ||
					     file_stat(filename)[3]));
  id->variables->date = id->variables->datelong[0..9];
  array tmp = remove_last_slash(dirname(id->not_query)) / "/";
  string mount = remove_last_slash(query("location"));
  
  id->variables->shortdirpath = tmp[sizeof(mount / "/")..] * "/";

  if (dir_info_short(filename) != "") {
    id->variables->dirname = dir_info_short(filename) || 
      id->variables->shortdirpath;
  } else {
    id->variables->dirname = id->variables->shortdirpath;
  }

  string indexFile = combine_path(filename, query("indexFileName"));
  
  if (in_edit_mode(id)) {
    return dir_edit(combine_path(filename,
				 query("dirInfoFilename")), id);
  }

  if (file_stat(indexFile)) {
    return Roxen.http_rxml_answer(Stdio.read_bytes(indexFile), id);
  } else {
    return directory_page(id);
  }
}

static int should_we_redirect_to_secure_server(RequestID id)
{
  string host = query("secureServerRedirectionHost");
  if (in_edit_mode(id) && !runs_ssl(id)) {
    return query("secureServerRedirection");
  }
  return 0;
}

static mapping show_not_real_file(string filename, string path, RequestID id)
{
  filename = remove_suffix(filename);
  if (!image_file_exists(filename) ||
      (image_hidden(filename)
      && !in_edit_mode(id))) {
    return 0;
  }
  if (!file_is_image(filename, id)) {
    return 0;
  }
  string imagename = query("location") + path[..sizeof(path)
					      - sizeof(query("suffix")) - 1];
  mapping fileStatsInDir = dir_stat(path, id);
  if (!fileStatsInDir) {
    return 0;
  }
  array(string) files = indices(fileStatsInDir);
  if (!in_edit_mode(id)) {
    files = remove_hidden_images(files, dirname(filename));
  }
  files = image_sort(dirname(filename), files, fileStatsInDir, id);
  set_previous_and_next_variables(files, path, id);

  id->variables->photo = imagename;
  id->variables->albumname = query("albumName");
  id->variables->photoname = image_desc_short(filename)
    || id->variables->photofilename;
  id->variables->pagename = id->variables->photofilename + query("suffix");
  id->variables->pagenameandquery = id->variables->pagename
    + query_part_of_url(id);
  id->variables->photodesc = image_desc_long(filename) || "";
  int utime = image_date(filename)
    || (fileStatsInDir[basename(filename)])[3];
  id->variables->date = stringtime(utime);
  id->variables->datelong = stringtime(utime);

  make_mid_size_image(imagename, filename, id);

  if (in_edit_mode(id) && is_jpeg_file(filename)) {
    return file_edit(imagename, filename, id);
  } else {
    return mid_image_page(id);
  }
}

string status()
{
  array s = the_cache->status();
  string cacheStatus = sprintf("<p><b>Images in cache:</b> %d<br />\n"
			       "<b>Cache size:</b> %s</p>",
			       s[0], Roxen.sizetostring(s[1]));
  if (supports_jpeg_transforms()) {
    return "This Pike version supports JPEG transforms. Good.<br />\n" +
    cacheStatus;
  } else {
    return "This Pike version does not support JPEG "
      "transforms, so external programs will be used instead:<br />\n" +
      program_status(getvar("wrjpgcomPath")) +
      program_status(getvar("rdjpgcomPath")) +
      program_status(getvar("jpegtranPath")) +
      cacheStatus;
  }
}

array(int)|Stat stat_file(string path, RequestID id)
{
  string file = combine_path(query("rootDir"), remove_first_slash(path));
  Stat statMap = file_stat(file);
  if (statMap && statMap->isreg && statMap->size >= 10 && is_jpeg_file(file))
      statMap->mtime = image_date(file) || exif_time(file) || statMap->mtime;
  return statMap;
}

static string stringtime(int unixtimeInt)
{
  mapping(string:int) tm = localtime(unixtimeInt);
  return
    sprintf("%04d:%02d:%02d %02d:%02d:%02d",
	    tm->year + 1900,
	    tm->mon + 1,
	    tm->mday,
	    tm->hour,
	    tm->min,
	    tm->sec);
}

static int(0..1) supports_video_decoding()
{
  //Always return zero for now
  return 0;
#if constant(_Ffmpeg.CODEC_ID_MJPEG)
  return 1;
#else
  return 0;
#endif
}

static int(0..1) supports_jpeg_transforms()
{
#if constant(Image.JPEG.ROT_90)
  return 1;
#else
  return 0;
#endif
}

static void swap_files(string filename, string oldFilename, string newFilename)
{
  mv(filename, oldFilename);
  mv(newFilename, filename);
}

static array(int(0..)) thumbnail_size(string realFile)
{
  Stdio.File fd = open(realFile, "r");
  int(0..) x,y;
  [x,y] = Dims.dims()->get(fd) || ({0,0});
  fd->close();
  if (x == 0 && y == 0) {
    return ({0,0});
  }
  float scale = (float) query("thumbnailSize") / max(x, y);
  if (scale > 1) {
    scale = 1.0;
  }
  int border = shadow_borders(THUMBNAIL_PASSEPARTOUT_SIZE);
  return ({ border + (int)(x*scale), border + (int)(y*scale) });
}

static mapping|void transform_image(string filename,
				    string oldFilename,
				    string newFilename,
				    RequestID id)
{
  if (supports_jpeg_transforms()) {
    int transform;
    switch (id->variables->trans) {
    case "rot90":
#if constant(Image.JPEG.ROT_90)
      transform = Image.JPEG.ROT_90;
#endif
      break;
    case "rot180":
#if constant(Image.JPEG.ROT_180)
      transform = Image.JPEG.ROT_180;
#endif
      break;
    case "rot270":
#if constant(Image.JPEG.ROT_270)
      transform = Image.JPEG.ROT_270;
#endif
      break;
    case "fliplr":
#if constant(Image.JPEG.FLIP_H)
      transform = Image.JPEG.FLIP_H;
#endif
      break;
    case "flipud":
#if constant(Image.JPEG.FLIP_V)
      transform = Image.JPEG.FLIP_V;
#endif
      break;
    case "nop":
      return;
    default:
      return aphoto_error("Strange error in image transformation "
			  "arguments from form. Operation missing!", id);
    }
#if constant(Image.JPEG.ROT_90)
    Stdio.write_file(newFilename,
		     Image.JPEG.encode(Stdio.read_bytes(filename),
				       (["transform":transform])));
#endif
  } else {
    Stdio.File newFile;
    object process;
    mixed error;
    
    string transformOperation = 0, transformArgument = 0;
    switch (id->variables->trans) {
    case "rot90":
      transformOperation = "-rotate";
      transformArgument = "90";
      break;
    case "rot180":
      transformOperation = "-rotate";
      transformArgument = "180";
      break;
    case "rot270":
      transformOperation = "-rotate";
      transformArgument = "270";
      break;
    case "fliplr":
      transformOperation = "-flip";
      transformArgument = "horizontal";
      break;
    case "flipud":
      transformOperation = "-flip";
      transformArgument = "vertical";
      break;
    case "nop":
      return;
    default:
      return aphoto_error("Strange error in image transformation "
			  "arguments from form. Operation missing!", id);
    };
    newFile = Stdio.File(newFilename, "cw");
    error = catch {
      process = Process.create_process(({ query("jpegtranPath"),
					  "-copy", "all",
					  transformOperation,
					  transformArgument,
					  filename}),
				       ([ "env":getenv(),
					  "stdout":newFile]));
    };
    if (error) {
      return aphoto_error("Couldn't execute jpegtran!", id);
    }
    process->wait();
  }
  if (!image_file_ok(newFilename)) {
    return aphoto_error("Something went very wrong when transforming "
			"the JPEG file. Exiting. Your original image"
			"file was not damaged.", id);
    
  }
  swap_files(filename, oldFilename, newFilename);
  if (image_file_ok(filename)) {
    rm(oldFilename);
  } else {
    return aphoto_error("Something went very wrong when transforming.\n"
			"the JPEG file. Old file was not deleted. It "
			"resides in " + oldFilename, id);
  }
}

static int unixtime(string timestring)
{
  // exif datetime is in YYYY:MM:DD HH:MM:SS format
  int yyyy, mon, dd, hh, mm, ss;
  if (sscanf(timestring, "%d:%d:%d %d:%d:%d", yyyy, mon, dd, hh, mm, ss) != 6)
    {
      return 0;
    }
  if (mon == 0)
    {
      return 0;
    }
  if (yyyy <= 1900 && yyyy >= 100)
    {
      return 0;
    }
  if (yyyy < 100) {      
    if (yyyy < 30) {
      yyyy += 2000;
    } else {
      yyyy += 1900;
    }
  }
  mapping(string:int) tm = ([
    "hour":hh,
    "mday":dd,
    "min":mm,
    "mon":mon - 1,
    "sec":ss,
    "year":yyyy - 1900
  ]);
  return mktime(tm);
}

static mapping|void write_new_jpeg_comment(string filename,
					   string oldFilename,
					   string newFilename,
					   string comment,
					   RequestID id)
{
  if (supports_jpeg_transforms()) {
#if constant(Image.JPEG.ROT_90)
    Stdio.write_file(newFilename,
		     Image.JPEG.encode(Stdio.read_bytes(filename),
				       (["comment":comment])));    
#endif
  } else {
    Stdio.File out = Stdio.File();
    Stdio.File newFile = Stdio.File(newFilename, "cw");
    
    object process;
    mixed error;
    error = catch {
      process = Process.create_process(({ query("wrjpgcomPath"),
					  "-replace",
					  filename }),
				       ([ "env":getenv() +
					  (["LC_ALL":query("LC_ALL")]),
					  "stdin":out->pipe(),
					  "stdout":newFile]));
    };
    if (error) {
      return aphoto_error("Couldn't execute wrjpgcom!", id);
    }
    out->write(comment);
    out->close();
    process->wait();
  }
}

static Image.image corner1 = Image.PNG.decode(DROPSHADOW_CORNER);
static Image.image corner2 = corner1->rotate_cw();
static Image.image corner3 = corner2->rotate_cw();
static Image.image corner4 = corner3->rotate_cw();

static Image.image make_border(Image.image in,
			       int border_width,
			       int r, int g, int b)
{
  Image.image border = Image.image(in->xsize() + border_width * 2,
				   in->ysize() + border_width * 2,
				   r, g, b);
  border->paste(in, border_width, border_width);
  return border;
}

static Image.image make_dropshadow(Image.image in, int width,
				   int border, int border1, int border2)
{
  if(width != in->xsize())
    in = in->scale(width, 0);
  in = make_border(in, border, 255,255,255);
  Image.image bg = Image.image(in->xsize() + border1 + border2,
			       in->ysize() + border1 + border2,
			       0x60,0x60,0x60);
  int cs = corner1->xsize();
  int cw = bg->xsize() - 2 * cs;
  int ch = bg->ysize() - 2 * cs;
  bg->paste(corner1, 0, 0);
  bg->paste(corner2, bg->xsize()-cs, 0);
  bg->paste(corner3, bg->xsize()-cs, bg->ysize()-cs);
  bg->paste(corner4, 0, bg->ysize()-cs);
  bg->paste(bg->copy(cs-1, 0, cs-1, cs-1)->scale(cw, cs), cs, 0);
  bg->paste(bg->copy(cs-1, ch+cs, cs-1, ch+cs+cs-1)->scale(cw, cs), cs, ch+cs);
  bg->paste(bg->copy(0, cs-1, cs-1, cs-1)->scale(cs, ch), 0, cs);
  bg->paste(bg->copy(cw+cs-1, cs-1, cw+cs+cs-1, cs-1)->scale(cs, ch),
	    cw+cs, cs);
  bg->paste(in, border1, border1);
  return bg;
}

class TagImgBorderZeroIfNetscape
{
  inherit RXML.Tag;
  constant name = "imgb0";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit RXML.Frame;
    array do_return(RequestID id) 
    {
      if (is_old_netscape(id)) {
	args->border = 0;
      }
      result = Roxen.make_tag("img", args, 1);
      return 0;
    }
  };
};

class TagPhotoDirectoryplugin
{
  inherit RXML.Tag;
  constant name = "emit";
  constant plugin_name = "photodir";

  array get_dataset(mapping args, RequestID id)
  {
    string d;
    if (args->directory)
      d = Roxen.fix_relative(args->directory, id);
    else
      d = dirname(id->not_query + "/");
    if (d[-1] != '/')
      d += "/";
    mapping a = id->conf->find_dir_stat(d, id);

    if (!a || !sizeof(a))
      return ({});

    mapping get_datum(string file)
    {
      Stdio.Stat st = a[ file ];

      mapping m =
	([
	  "name":file,
	  "filename":file,
	  "dirname":d,
	  "dirinfoshort":0,
	  "dirinfolong":0,
	  "infoorfilename":"",
	  "path":combine_path(d, file),
	  "atime-unix":st->atime,
	  "mtime-unix":st->mtime,
	  "mtime-iso":Roxen.strftime("%Y-%m-%d", st->mtime),
	  "atime-iso":Roxen.strftime("%Y-%m-%d", st->atime),
	  "mode-int":st->mode,
	  "filesize":st->size,
	  "mode":st->mode_string,
	  "mtime-full":Roxen.strftime("%Y-%m-%d %H:%M:%S", st->mtime),
	]);

      if (args->strftime)
	{
	  m->mtime = Roxen.strftime(args->strftime, st->mtime);
	  m->atime = Roxen.strftime(args->strftime, st->atime);
	}  else {
	  m->mtime = m["mtime-iso"];
	  m->atime = m["atime-iso"];
	}

      if (st->isdir)
	{
	  m->size = "0";
	  m->type = "directory";
	  m->dirinfoshort = dir_info_short(id->conf->real_file(m->path, id));
	  m->dirinfolong = dir_info_long(id->conf->real_file(m->path, id));
	  m["size"] = "";
	  m["type-img"] = "internal-gopher-menu";
	} else {
	  m->type = id->conf->type_from_filename(file);
	  m->size = Roxen.sizetostring(st->size);
	  m["type-img"] = Roxen.image_from_type(m->type);
	}

      return m;
    };

    array files = indices(a);

    // remove all non directory files
    // and hidden directories
    files = remove_nondir_and_hidden(a, d, id);

    if (args->glob) {
      array tmp = ({});
      foreach(args->glob/",", string g)
        tmp |= glob(g, files);
      files = tmp;
    }

    array res = map(files, get_datum);

    if (args->type)
      {
	array tmp = ({});
	foreach(res, mapping a)
	  {
	    foreach(args->type/",", string g)
	      if (glob(g, a->type))
		{
		  tmp += ({ a });
		  break;
		}
	  }
      }

    if (!args->sort) {
      args->sort = query("dirSortOrder");
    }
    
    // add / to end of path. Works since only directories
    foreach (res, mapping a) {
      a["filename"] = a["filename"] + "/";
      a["name"] = a["name"] + "/";
      a["path"] = a["path"] + "/";      
      if (a["dirinfoshort"] != "" && a["dirinfoshort"] != 0)
	a["infoorfilename"] = a["dirinfoshort"];
      else
	a["infoorfilename"] = a["filename"];
    }
    return res;
  }
}

class TagThumbnailsPlugin {
  inherit RXML.Tag;
  constant name = "emit";
  constant plugin_name = "thumbnails";

  array get_dataset(mapping args, RequestID id)
  {
      string d = id->not_query;
      string realfile = id->conf->real_file(d, id);
      if (file_stat(realfile)->isdir)
	if (d[-1] != '/') {
	  d += "/";
	}

      mapping a = id->conf->find_dir_stat(d, id);

      if (!a || !sizeof(a))
	return ({});

      array files = indices(a);
      array hideFiles = query("hideFiles");
      foreach(hideFiles, string hideFile)
	if (a[hideFile])
	  return ({});
      array(string) tmp = ({});
      foreach (files, string f) {
	if (((id->conf->type_from_filename(f) / "/") [0]) == "image" ||
	    ((supports_video_decoding() &&
	      (id->conf->type_from_filename(f) / "/") [0]) == "video")) {
	  tmp += ({f});
	}
      }
      files = image_sort(realfile, tmp, a, id);
      if (!in_edit_mode(id)) {
	files = remove_hidden_images(files, realfile);
      }
      
      string path = "";
      array(mapping) res = ({});
      mapping m;
      foreach (files, string filename) {
	path = d + filename;
	int(0..) x, y;
	[x,y] = thumbnail_size(id->conf->real_file(path, id));
	mapping m =
	  ([
	    "filename":filename,
	    "path":path,
	    "thumbnail":query_absolute_internal_location(id)
	    + the_cache->store(get_my_args(([
	      "imagetype":"thumbnail",
	      "file":id->conf->real_file(path, id),
	      "query":query_part_of_url(id),
	      "format":"jpg"]), id), id),
	    "title":image_desc_short(id->conf->real_file(path, id)) || filename,
	    "width":(string) x,
	    "height":(string) y,
	    "href":path + query("suffix")
	  ]);
	res += ({ m });
      }
      return res;
  }
}

TAGDOCUMENTATION;

#ifdef manual
constant tagdoc=([
  "emit#photodir":({ "<desc type='plugin'><p><short>"
		     "Entitybased directory name list"
		     "</short>.</p></desc>",
		     
		     ([
		       "&_.name;":
		       "<desc type='entity'><p>Returns the filename.</p>"
		       "</desc>",
		       
		       "&_.filename;":
		       "<desc type='entity'><p>Returns the filename.</p>"
		       "</desc>",
		       
		       "&_.dirname;":"<desc type='entity'><p>Returns the "
		       "path up to filname.</p></desc>",

		       "&_.path;":"<desc type='entity'><p>Returns the path, "
		       "including filename.</p></desc>",
		       
		       "&_.dirinfoshort;":"<desc type='entity'><p>Returns a "
		       "short directory description.</p></desc>",
		       
		       "&_.dirinfolong;":"<desc type='entity'><p>Returns a "
		       "long directory description.</p> "
		       "</desc>"
		     ])}),
  
  "emit#thumbnails":({ "<desc type='plugin'><p><short>"
		       "Entitybased thumbnail link and image list"
		       "</short>.</p></desc>",
		       
		       ([
			 "&_.filename;":
			 "<desc type='entity'><p>Returns the filename.</p>"
			 "</desc>",

			 "&_.path;":
			 "<desc type='entity'><p>Returns the full path.</p>"
			 "</desc>",

			 "&_.thumbnail;":
			 "<desc type='entity'><p>Returns a path to the cached "
			 "thumbnail.</p>"
			 "</desc>",

			 "&_.title;":
			 "<desc type='entity'><p>Returns a short description "
			 "of the image, or the filename if no short "
			 "description exists.</p>"
			 "</desc>",

			 "&_.width;":
			 "<desc type='entity'><p>Returns the width of the "
			 "generated thumbnail.</p>"
			 "</desc>",

			 "&_.height;":
			 "<desc type='entity'><p>Returns the height of the "
			 "generated thumbnail.</p>"
			 "</desc>",

			 "&_.href;":
			 "<desc type='entity'><p>Returns a link to the "
			 "real image.</p>"
			 "</desc>",
		     ])}),
  "imgb0":"<desc tag='tag'><p><short>img wrapper tag</short>. "
  "The imgb0-tag is a wrapper for the common img tag, which "
  "adds an additional border=\"0\" to the arguments if an "
  "old Netscape browser asks for the page. Useful, since no "
  "Netscape&nbsp;4.X understands the CSS attribute border: none "
  "</p></desc>",
]);
#endif

static constant THUMBNAIL_PASSEPARTOUT_SIZE = 7;
static constant PHOTO_PASSEPARTOUT_SIZE = 20;
static constant SHADOW_BORDER_TOPLEFT = 7;
static constant SHADOW_BORDER_BOTTOMRIGHT = 11;

static constant IMAGE_DESC_SHORT = "image-desc-short";
static constant IMAGE_DESC_LONG = "image-desc-long";
static constant IMAGE_DATE = "image-date";
static constant IMAGE_HIDDEN = "is-hidden";
static constant DIR_INFO_SHORT = "dir-desc-short";
static constant DIR_INFO_LONG = "dir-desc-long";
static constant DIR_DATE = "dir-date";
static constant MODULE_NAME = "aphoto";
static constant MODULE_URL = "http://www.lysator.liu.se/~norrby/aphoto.html";
static constant DATA_EDITING_PERMISSION = "Aphoto data editing";

static constant PREFETCH_HTML =
"<if variable=\"form.prefetchleft\">"
" <link rel=\"prefetch\" href=\"&form.prefetchleft;\"/>\n</if>"
"<if variable=\"form.prefetchright\">"
" <link rel=\"prefetch\" href=\"&form.prefetchright;\"/>\n</if>";

static constant PAGE_HEADER =
#"<?xml version=\"1.0\"?>
<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"
\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">
<html>
<head>
 <link rel=\"stylesheet\" type=\"text/css\" href=\"&form.internal;aphoto.css\""
+ " title=\"default aphoto style\" />\n"
+ PREFETCH_HTML;

static constant PAGE_FOOTER =
#"</html>\n";

static constant APHOTO_VERSION_HTML = "<span class=\"aphotoversion\"><a href=\""
+ MODULE_URL + "\">"
+ MODULE_NAME
+ "</a>&nbsp;v&nbsp;&form.aphotoversion;</span>";

static constant APHOTO_CSS = #"
body {
  margin: 0;
  padding: 0;
  border: none;
  background-color: white;
  color: black;
}
h1 {
  margin: 0;
  padding: 0.1em 0.1em 0.1em 10px;;
  font-family: sans-serif;
  font-weight: normal;
}
div.dir {
  padding: 0px 7px 0px 0px;
  margin: 0 0 0 1em;
}
img.dirlink, a.dirlink {
  vertical-align: middle;
  border-style: none;
}
input, textarea {
  background: lightblue;
  border-style: none;
  font-size: 100%;
}
table.outer {
  width: 100%;
  border-style: none;
  border-collapse: collapse;
}
img.thumbnail {
  border-style: none;
}
td#topleft {
  width: 110px;
  padding-left: 1em;
  padding-right: 0.5em;
  border-right: solid black 1px;
  border-bottom: solid black 1px;
  text-align: center;
  vertical-align: middle;
}
td#topright {
  border-left: solid black 1px;
  border-bottom: solid black 1px;
  vertical-align: middle;
}
td#secondhighestleft {
  width: 110px;
  padding-left: 1em;
  padding-right: 0.5em;
  border-right: solid black 1px;
  text-align: left;
  height: 10px;
}
td#secondhighestright {
  border-left: solid black 1px;
  padding: 0.5em 1em 0.2em 15px;
  margin-bottom: 1em;
  height: 10px;
  vertical-align: top;
}
td#middleright {
  border-left: solid black 1px;
  padding-left: 5px;
  vertical-align: top;
} 
td#middleleft {
  width: 110px;
  padding-left: 1em;
  padding-right: 0.5em;
  border-right: solid black 1px;
  vertical-align: top;
  text-align: center;
}
td#secondlowestright {
  border-left: solid black 1px;
  padding: 0.5em 0 0.5em 0.2em;
  margin-top: 1em;
} 
td#secondlowestleft {
  width: 110px;
  padding-left: 1em;
  padding-right: 0.5em;
  text-align: center;
}
td#bottomright {
  border-left: solid black 1px;
  padding: 0.5em 0 0.3em 1em;
  vertical-align: bottom;
} 
td#bottomleft {
  width: 110px;
  padding: 0.5em 0.5em 0.3em 1em;
  border-right: solid black 1px;
  text-align: center;
  vertical-align: bottom;
}
img#photo {
  border: none;
  padding-left: 10px;
}
";

static constant PREVIOUS_NEXT_ARROWS_HTML = 
"<if variable=\"form.previous\"><a href=\"&form.previous;\">"
"<imgb0 class=\"dirlink\" src=\"&form.internal;arrow-left\" alt=\"&lt;&lt;\" />"
"</a></if><else>"
"<imgb0 class=\"dirlink\" src=\"&form.internal;arrow-left-nonactive\" "
"alt=\"&lt;&lt;\" /></else><if variable=\"form.parent\">"
"<a href=\"&form.parent;\">"
"<imgb0 class=\"dirlink\" src=\"&form.internal;arrow-up\" alt=\"^^\" />"
"</a></if><else>"
"<imgb0 class=\"dirlink\" src=\"&form.internal;arrow-up-nonactive\" "
"alt=\"^^\" /></else><if variable=\"form.next\"><a href=\"&form.next;\">"
"<imgb0 class=\"dirlink\" src=\"&form.internal;arrow-right\" "
"alt=\"&gt;&gt;\" /></a></if><else><imgb0 class=\"dirlink\" "
"src=\"&form.internal;arrow-right-nonactive\" alt=\"&gt;&gt;\" /></else>";

static constant BOTTOM_ROW_HTML =
#"  <tr>
   <td id=\"bottomleft\">
    <i><a href=\"http://www.lysator.liu.se/~norrby/aphoto.html\">aphoto</a>" +
"&nbsp;v&nbsp;&form.aphotoversion;</i>" + #"
   </td>
   <td id=\"bottomright\">&form.date;&#8212;" +
"<a href=\"./&form.pagename;?edit=true\"><i>&lt;edit&gt;</i></a></td></tr>";

static constant BOTTOM_ROW_EDIT_HTML =
#"    <tr>
     <td id=\"bottomleft\">
      <a href=\"./\"><i>&lt;exit&nbsp;edit&gt;</i></a>
     </td>
     <td id=\"bottomright\">
      <input name=\"new_date\" size=\"32\" value=\"&form.datelong;\"/>
     </td>
    </tr>";

static constant THUMBNAILS_HTML =
#"<emit source=\"thumbnails\">
    <a href=\"&_.href;\"><imgb0 class=\"thumbnail\" src=\"&_.thumbnail;\" "
"alt=\"\" width=\"&_.width;\" height=\"&_.height;\" title=\"&_.title;\" />"
"</a></emit>";

static constant DIRECTORIES_HTML =
#"<emit source=\"photodir\">
    <div class=\"dir\"><span><a class=\"dirlink\" href=\"&_.path;\">" +
"<imgb0 class=\"dirlink\" src=\"&form.internal;folder\" height=\"32\" "
"width=\"32\" alt=\"[*]\" /></a></span>&nbsp;<a class=\"dirlink\" "
"href=\"&_.path;\">&_.infoorfilename;</a></div></emit>";

static constant DIRINFO_HTML =
#"    <div><i><autoformat>&form.longdirinfo;</autoformat></i></div>";

static constant SITENAV_HTML =
#"<if variable=\"form.previous\"> <link rel=\"prev\" href=\"&form.previous;\"/>
 <link rel=\"first\" href=\"&form.first;\"/>
</if><if variable=\"form.start\"> <link rel=\"start\" href=\"&form.start;\"/>
</if><if variable=\"form.next\"> <link rel=\"next\" href=\"&form.next;\"/>
 <link rel=\"last\" href=\"&form.last;\"/>
</if><if variable=\"form.parent\"> <link rel=\"parent\" href=\"&form.parent;\"/>
</if>";

static constant DEFAULT_THUMBNAIL_PAGE = 
PAGE_HEADER +
SITENAV_HTML +
#" <title>&form.albumname;<if sizeof=\"form.dirname > 0\">" +
#": &form.dirname;</if></title>
</head>
 <body>
 <table class=\"outer\">
  <tr>
   <td id=\"topleft\">" + PREVIOUS_NEXT_ARROWS_HTML + #"
   </td>
   <td id=\"topright\"><h1>&form.albumname;<if sizeof=\"form.dirname > 0\">" +
#"&nbsp;&#8212;&nbsp;&form.dirname;</if></h1></td>
  </tr>
  <tr>
   <td id=\"secondhighestleft\"></td>
   <td id=\"secondhighestright\">" + DIRINFO_HTML + #"
   </td>
  </tr>
  <tr>
   <td id=\"middleleft\"></td>
   <td id=\"middleright\">" + DIRECTORIES_HTML + #"
   </td>
  </tr>
  <tr>
   <td id=\"secondlowestleft\"></td>
   <td id=\"secondlowestright\">" + THUMBNAILS_HTML + #"
   </td>
  </tr>
" + BOTTOM_ROW_HTML + #"
 </table>
</body>
" + PAGE_FOOTER;

static constant DEFAULT_THUMBNAIL_EDIT_PAGE =
PAGE_HEADER +
SITENAV_HTML +
#" <title>&form.albumname;<if sizeof=\"form.dirname > 0\">" +
#": &form.dirname;</if></title>
</head>
<body>
 <form action=\"./&form.pagenameandquery;\" method=\"post\">
  <div>
   <input type=\"hidden\" name=\"full_path_in_real_file_system\" "
    #"value=\"&form.full_path;\" />
   <input type=\"hidden\" name=\"aphoto_editing_data\" value=\"posted\" />
   <input type=\"hidden\" name=\"short_dir_path\" "
   #"value=\"&form.shortdirpath;\" />  <table class=\"outer\">
    <tr>
     <td id=\"topleft\">" + PREVIOUS_NEXT_ARROWS_HTML + #"
     </td>
     <td id=\"topright\"><h1>&form.albumname;" +
  #"&nbsp;&#8212;&nbsp;<input name=\"new_dir_info_short\" "
  #"value=\"&form.dirname;\"/></h1></td>
    </tr>
    <tr>
     <td id=\"secondhighestleft\"></td>
     <td id=\"secondhighestright\">
      <textarea name=\"new_dir_info_long\" cols=\"64\" rows=\"4\">"
                   + #"&form.longdirinfo:none;</textarea>
     </td>
    </tr>
    <tr>
     <td id=\"middleleft\">
      <input type=\"submit\" name=\"submit_and_stay\" value=\"OK\" />
     </td>
     <td id=\"middleright\">" + DIRECTORIES_HTML + #"
     </td>
    </tr>
    <tr>
     <td id=\"secondlowestleft\"></td>
     <td id=\"secondlowestright\"><emit source=\"thumbnails\">
      <a href=\"&_.href;?edit=true\"><imgb0 class=\"thumbnail\" "
      "src=\"&_.thumbnail;\" alt=\"\" "
      "width=\"&_.width;\" height=\"&_.height;\" title=\"&_.title;\" /></a>"
      #"</emit>
     </td>
    </tr>" + BOTTOM_ROW_EDIT_HTML + #"
   </table>
  </div>
 </form>
</body>
" + PAGE_FOOTER;

static constant DEFAULT_PHOTO_PAGE = 
PAGE_HEADER +
SITENAV_HTML +
#" <title>&form.albumname;: &form.photoname;</title>
</head>
<body>
 <table class=\"outer\">
  <tr>
   <td id=\"topleft\">" + PREVIOUS_NEXT_ARROWS_HTML + #"
   </td>
   <td id=\"topright\"><h1>&form.photoname;</h1></td>
  </tr>
  <tr>
   <td id=\"secondhighestleft\"></td>
   <td id=\"secondhighestright\"><i>&form.photodesc;</i></td>
  </tr>
</table>
<div><a href=\"&form.photo;\"><imgb0 id=\"photo\"
        src=\"&form.midiphoto;\"
        width=\"&form.midiphotowidth;\"
        height=\"&form.midiphotoheight;\"
        title=\"[&form.fullimagewidth;&form.fullimageheight;], "
             + #"&form.fullimagesizekbytes; kB\"
	alt=\"\" /></a></div>
 <table class=\"outer\">
" + BOTTOM_ROW_HTML + #"
 </table>
</body>
" + PAGE_FOOTER;

static constant DEFAULT_PHOTO_EDIT_PAGE = 
PAGE_HEADER +
SITENAV_HTML +
#" <title>Editing: &form.photoname;</title>
</head>
<body>
 <form action=\"./&form.pagenameandquery;\" method=\"post\">
  <div>
   <input type=\"hidden\" name=\"aphoto_editing_data\" value=\"posted\" />
   <input type=\"hidden\" name=\"full_path_in_real_file_system\" " 
             + #"value=\"&form.full_path:none;\" />
   <input type=\"hidden\" name=\"previous\" value=\"&form.previous;\" />
   <input type=\"hidden\" name=\"next\" value=\"&form.next;\" />
   <table class=\"outer\">
    <tr>
     <td id=\"topleft\">" + PREVIOUS_NEXT_ARROWS_HTML + #"
     </td>
     <td id=\"topright\"><h1><input size=\"32\" value=\"&form.photoname;\" "
          + #"name=\"new_image_desc_short\" /></h1></td>
    </tr>
    <tr>
     <td id=\"secondhighestleft\">
      Transform (lossless)<br />
      <input type=\"radio\" name=\"trans\" value=\"nop\" checked=\"checked\"/>"
                 +#"<i>no transf.</i><br />
      <input type=\"radio\" name=\"trans\" value=\"rot90\"/>Rot 90<br />
      <input type=\"radio\" name=\"trans\" value=\"rot180\"/>Rot 180<br />
      <input type=\"radio\" name=\"trans\" value=\"rot270\"/>Rot 270<br />
      <input type=\"radio\" name=\"trans\" value=\"flipud\"/>Flip U/D<br />
      <input type=\"radio\" name=\"trans\" value=\"fliplr\"/>Flip R/L<br />
     </td>
     <td id=\"secondhighestright\"><textarea name=\"new_image_desc_long\" " +
             #"rows=\"5\" cols=\"64\">&form.photodesc;</textarea>
     </td>
    </tr>
    <tr>
     <td id=\"middleleft\"></td>
     <td id=\"middleright\">&nbsp;
      <input type=\"submit\" name=\"submit_and_previous\" value=\"&lt;&#8212;" 
          + #"OK\" />
      <input type=\"submit\" name=\"submit_and_stay\" value=\"OK\" />
      <input type=\"submit\" name=\"submit_and_next\" value=\"OK&#8212;&gt;\" />
      <if variable=\"form.hidden\"><input type=\"submit\" name=\"unhide\" "
      #"value=\"Make visible\" />
      </if><else><input type=\"submit\" name=\"hide_and_next\" "
      "value=\"Hide&#8212;&gt;\" "
      #"/></else>
     </td>
    </tr>
   </table>
   <div><a href=\"&form.photo;\"><imgb0 id=\"photo\"
           src=\"&form.midiphoto;\"
           width=\"&form.midiphotowidth;\"
           height=\"&form.midiphotoheight;\"
           title=\"[&form.fullimagewidth;&form.fullimageheight;], "
                + #"&form.fullimagesizekbytes; kB\"
           alt=\"\" /></a></div>
   <table class=\"outer\">" + BOTTOM_ROW_EDIT_HTML + #"
   </table>
  </div>
 </form>
</body>
" + PAGE_FOOTER;

static constant FOLDER = "GIF87a \0 \0\0\0\0\0\200\0\0\0\200\0\200\200\0\0\0\200\200\0\200\0\200\200\231f3\231f3\0\231\231\231\231\231f\231""3\231\0fff\231fff3f\0333\2313f333\0\0\0\231\0f\0""3\231ff3\0\231f3\0\231\231\231\231\231f\231""3\231\0fff\231fff3f\0333\2313f333\0\0\0\0\231\0f\0""3\0\0\231\231\231\231\231f\2313\231\0\231\231\231\231\231f\2313\231\0\231\231\231\231\231\231\231\231\231f\231\231""3\231\231\0\231f\231f\231f\231\231ff\231f3\231f\0\231""3\231""3\231""3\231\231""3f\231""33\231""3\0\231\0\231\0\231\0\231\231\0f\231\0""3\231\0\0fff\231fff3f\0fff\231fff3f\0f\231f\231f\231\231f\231ff\231""3f\231\0ffffff\231fffff3ff\0f3f3f3\231f3ff33f3\0f\0f\0f\0\231f\0ff\0""3f\0\0""333\231""3f333\0""333\231""3f333\0""3\2313\2313\231\231""3\231f3\231""33\231\0""3f3f3f\231""3ff3f33f\0""333333\231""33f33333\0""3\03\03\0\231""3\0f3\0""33\0\0\0\0\231\0f\03\0\0\0\231\0f\03\0\0\0\231\0\231\0\231\231\0\231f\0\231""3\0\231\0\0f\0f\0f\231\0ff\0f\0\0f3\0""3\0""3\0""3\231\0""3f\0""33\0""3\0\0\0\0\0\231\0\0f\0\0""3\0\0\0\0\0\0\210\0\0w\0\0U\0\0D\0\0\"\0\0\0\0\0\0\0\0\0\210\0\0w\0\0U\0\0D\0\0\"\0UUUwwwwwwDDD\"\"\"\21\21\21\0\0w\0\0U\0\0D\0\0\"𠠤\200\200\200\0\0\0\0\0\0\0\0\0!\4\1\0\0\1\0,\0\0\0\0 \0 \0\0\b\0\3\b\34H\203\b\23*\\Ȱ\207\20#J\234h\20\200\213\30)\6\0Х\217\0&\2GI\222\30S\34\b\200\227""0c\234\31r\230r_\226""7]\14*g\236\7[gK\31\7va\7`\213MI\227""6\t4˩U72UJR*Jj\25Xer\225!)\3`۾\205\25#]N\25V\213\177Zd./ahS&\6\14\200\235c\220\17\3\234آ{\rys\213\35\7c\216t\30}Z\6z\235m\210K/nL[*\203sT=\226m\31es\31\207\207l\226K}{]4UysͥmZ<p)A\37""8I>\7Ͽ'\200V\21\201\b&\202\13\6\4\0;";

static constant ARROW_LEFT = "GIF89a\30\0\30\0\221\0\0\0\0\0\0\0!\4\1\0\0\2\0,\0\0\0\0\30\0\30\0@\2+\224\217\33\2\234\213v/}Fj\b&ʶ.ĩhZh\227\206\17\24\0\0;";

static constant ARROW_LEFT_NONACTIVE = "GIF87a\30\0\30\0\0\0\0\0\0,\0\0\0\0\30\0\30\0\0\2\26\204\217\17\234\213޼\17\206H\226S\1\0;";

static constant ARROW_RIGHT = "GIF89a\30\0\30\0\221\0\0\0\0\0\0\0!\4\1\0\0\2\0,\0\0\0\0\30\0\30\0@\2*\224\217\31ߢ\201\207\17^\13Gn\204\235ʶnة\31R֩`r\14\177(\0\0;";

static constant ARROW_RIGHT_NONACTIVE = ARROW_LEFT_NONACTIVE;

static constant ARROW_UP = "GIF89a\30\0\30\0\0\0\0\0\0!\4\1\0\0\1\0,\0\0\0\0\30\0\30\0\0\2+\214\217\0\14\221s\21Yս}_ad\235\222\32i\33ҧM\17\14\n\0;";

static constant ARROW_UP_NONACTIVE = ARROW_LEFT_NONACTIVE;

static constant DROPSHADOW_CORNER = "\211PNG\r\n\32\n\0\0\0\rIHDR\0\0\0\32\0\0\0\32\b\2\0\0\0&(\231\0\0\2""1IDATx\223k\32\1\34\205OmUUF\24<4APA\211A\20*A\211\27\205\200\7\30\227\34Q\211\6&.FZj:ִc\227\14߼~o\204|\226\227\177\224G\20\17>/i\217[KXc\211 \21""1\233ͦ\207\205o'a\224\b\202³dr}}=\32\215-v\13\t\202\205#\21\4e8\34zNnl!6k<\36\3\2qqqq~~h4ji\211/2\21`v\1JR.\227\13\205B>\237\177\215\22V\213\214\202l6jT\215l6\233dt2\231L$\22/6JXY#&}\221j:99rT*\26\213Ep8|tt\24\n\205\236o\22Gk5\"cX\204\216\217\217y)\20\b|Cl\204UR6\214\210/X`\\.n\236Jk\211\23\217@RxzzJ_x\234h^tBZfd2>\221\r\2168\2\233\216tO_d\27v,\26\213h:\235~;{ʾ\4w\214D\"~p\13\226`jjZR=\222\r\216_\204;U&#\27\31hx[T*\24\212\207ZǱl(\200(\213\214\230\202%\227e2\3i\206\23qL\4\348p\205""5|\211\22217v\27\35\203#\213Ⱥ';\35\177\3?\b\205m@4\3\0\0\0\0IENDB`\202";
